#
# coding=utf-8
"""events - Event models for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# pylint: disable-msg=C0121

import contextlib
import csv
import datetime
import glob
import inspect
import operator
import os

try:
    import cPickle as pickle
except ImportError:  # Python 3, and 2.x without speedy helper
    import pickle  # NOQA

import click

from . import compat
from . import utils


class TaskNotRunningError(utils.RdialError):

    """Exception for calling mutators when a task is not running."""

    pass


class TaskRunningError(utils.RdialError):

    """Exception for starting task when a task is already running."""

    pass


class TaskNotExistError(utils.RdialError):

    """Exception for attempting to operate on a non-existing task."""

    pass


class Event(object):

    """Base object for handling database event."""

    def __init__(self, task, start=None, delta=None, message=''):
        """Initialise a new ``Event`` object.

        :param str task: Task name to tracking
        :param datetime.datetime start: Start time for event
        :param datetime.timedelta delta: Duration for event
        :param str message: Message to attach to event
        """
        self.task = task
        if isinstance(start, datetime.datetime):
            if not start.tzinfo:
                raise ValueError('Must not be a naive datetime %r' % start)
            self.start = start
        else:
            self.start = utils.parse_datetime(start)
        if isinstance(delta, datetime.timedelta):
            self.delta = delta
        else:
            self.delta = utils.parse_delta(delta)
        self.message = message

    @compat.mangle_repr_type
    def __repr__(self):
        """Self-documenting string representation.

        :rtype: :obj:`str`
        :return: Event representation suitable for :func:`eval`
        """
        return 'Event(%r, %r, %r, %r)' \
            % (self.task, utils.format_datetime(self.start),
               utils.format_delta(self.delta), self.message)

    def writer(self):
        """Prepare object for export.

        :rtype: :obj:`dict`
        :return: Event data for object storage
        """
        return {
            'start': utils.format_datetime(self.start),
            'delta': utils.format_delta(self.delta),
            'message': self.message,
        }

    def running(self):
        """Check if event is running.

        :rtype: :obj:`str`
        :return: Event name, if running
        """
        if self.delta == datetime.timedelta(0):
            return self.task
        return False

    def stop(self, message=None, force=False):
        """Stop running event.

        :param str message: Message to attach to event
        :param bool force: Re-stop a previously stopped event
        :raise TaskNotRunningError: Event not running
        """
        if not force and self.delta:
            raise TaskNotRunningError('No task running!')
        self.delta = utils.utcnow() - self.start
        self.message = message
FIELDS = inspect.getargspec(Event.__init__)[0][2:]


class Events(list):

    """Container for database events."""

    def __init__(self, iterable=None, backup=True):
        """Initialise a new ``Events`` object.

        :param list iterable: Objects to add to container
        :param bool backup: Whether to create backup files
        """
        super(Events, self).__init__(iterable if iterable else [])
        self.backup = backup
        self._dirty = []

    @compat.mangle_repr_type
    def __repr__(self):
        """Self-documenting string representation.

        :rtype: :obj:`str`
        :return: Events representation suitable for :func:`eval`
        """
        return 'Events(%s)' % super(self.__class__, self).__repr__()

    @property
    def dirty(self):
        """Tasks requiring sync against storage."""
        return self._dirty

    @dirty.setter
    def dirty(self, value):
        """Mark task as needing sync.

        :param str value: Task to mark as dirty
        """
        if not value in self._dirty:
            self._dirty.append(value)

    @dirty.deleter
    def dirty(self):
        """Mark dirty queue as flushed."""
        self._dirty = []

    @staticmethod
    def read(directory, backup=True, write_cache=True):
        """Read and parse database.

        Assume a new :obj:`Events` object should be created if the file is
        missing

        :param str directory: Location to read database files from
        :param bool backup: Whether to create backup files
        :param bool write_cache: Whether to write cache files
        :rtype: :obj:`Events`
        :returns: Parsed events database
        """
        if not os.path.exists(directory):
            return Events(backup=backup)
        events = []
        cache_dir = os.path.join(utils.xdg_cache_location(),
                                 directory.replace('/', '_'))
        if write_cache and not os.path.isdir(cache_dir):
            os.makedirs(cache_dir)
        for file in glob.glob('%s/*.csv' % directory):
            task = os.path.basename(file)[:-4]
            cache_file = os.path.join(cache_dir, task) + '.pkl'
            evs = None
            if os.path.exists(cache_file) and utils.newer(cache_file, file):
                try:
                    evs = pickle.load(open(cache_file))
                except pickle.UnpicklingError:
                    pass
            if evs is None:
                evs = [Event(task, **d)
                       for d in list(csv.DictReader(open(file)))]
                if write_cache:
                    pickle.dump(evs, open(cache_file, 'w'), -1)
            events.extend(evs)
        return Events(sorted(events, key=operator.attrgetter('start')))

    def write(self, directory):
        """Write database file.

        :param str directory: Location to write database files to
        """
        if not os.path.isdir(directory):
            os.makedirs(directory)

        for task in self.dirty:
            task_file = '%s/%s.csv' % (directory, task)
            events = self.for_task(task)
            with click.utils.LazyFile(task_file, 'w', atomic=True) as temp:
                if compat.PY2:
                    writer = csv.DictWriter(temp, FIELDS, lineterminator='\n')
                else:
                    writer = csv.DictWriter(temp, FIELDS,
                                            dialect=csv.unix_dialect,
                                            quoting=csv.QUOTE_MINIMAL)
                # Can't use writeheader, it wasn't added until 2.7.
                writer.writerow(dict(zip(FIELDS, FIELDS)))
                for event in events:
                    writer.writerow(event.writer())
                if self.backup and os.path.exists(task_file):
                    os.rename(task_file, '%s~' % task_file)
        del self.dirty

    def tasks(self):
        """Generate a list of tasks in the database.

        :rtype: :obj:`list` of :obj:`str`
        :return: Names of tasks in database
        """
        return sorted(set(event.task for event in self))

    def last(self):
        """Return current/last event.

        This handles the empty database case by returning ``None``

        :rtype: `Event`
        :return: Last recorded event
        """
        if len(self) > 0:
            return self[-1]
        else:
            return None

    def running(self):
        """Check if an event is running.

        We return the running task, if one exists, for easy access.

        :rtype: `Event`
        :return: Running event, if an event running
        """
        last = self.last()
        return last.running() if last else False

    def start(self, task, new=False, start=''):
        """Start a new event.

        :param str task: Task name to tracking
        :param str start: ISO-8601 start time for event
        :raise TaskRunningError: An event is already running
        """
        if not new and task not in self.tasks():
            raise TaskNotExistError("Task %s does not exist!  Use `--new' to "
                                    'create it' % task)
        running = self.running()
        if running:
            raise TaskRunningError('Running task %s!' % running)
        self.append(Event(task, start))
        self.dirty = task

    def stop(self, message=None, force=False):
        """Stop running event.

        :param str message: Message to attach to event
        :param bool force: Re-stop a previously stopped event
        :raise TaskNotRunningError: No task running!
        """
        if not force and not self.running():
            raise TaskNotRunningError('No task running!')
        self.last().stop(message, force)
        self.dirty = self.last().task

    def filter(self, filt):
        """Apply filter to events.

        :param func filt: Function to filter with
        :rtype: :obj:`Events`
        :return: Events matching given filter function
        """
        return Events(filter(filt, self))

    def for_task(self, task):
        """Filter events for a specific task.

        :param str task: Task name to filter on
        :rtype: :obj:`Events`
        :return: Events marked with given task name
        """
        return self.filter(lambda x: x.task == task)

    def for_date(self, year, month=None, day=None):
        """Filter events for a specific date.

        :param int year: Year to filter on
        :param int month: Month to filter on, or :obj:`None`
        :param int day: Day to filter on, or :obj:`None`
        :rtype: :obj:`Events`
        :return: Events occurring within specified date
        """
        events = self.filter(lambda x: x.start.year == year)
        if month:
            events = events.filter(lambda x: x.start.month == month)
        if day:
            events = events.filter(lambda x: x.start.day == day)
        return events

    def for_week(self, year, week):
        """Filter events for a specific ISO-8601 week.

        :param int year: Year to filter events on
        :param int week: ISO-8601 month number to filter events on
        :rtype: :obj:`Events`
        :return: Events occurring in given ISO-8601 week
        """
        start, end = utils.iso_week_to_date(year, week)
        return self.filter(lambda x: start <= x.start.date() < end)

    def sum(self):
        """Sum duration of all events.

        :rtype: :obj:`datetime.timedelta`
        :return: Sum of all event deltas
        """
        return sum(map(operator.attrgetter('delta'), self),
                   datetime.timedelta(0))

    @staticmethod
    @contextlib.contextmanager
    def context(directory, backup=True, write_cache=True):
        """Convenience context handler to manage reading and writing database.

        :param str directory: Database location
        :param bool backup: Whether to create backup files
        :param bool write_cache: Whether to write cache files
        """
        events = Events.read(directory, backup, write_cache)
        yield events
        if events.dirty:
            events.write(directory)



#
# coding=utf-8
"""test_attrdict - Test AttrDict support"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest import TestCase

from expecter import expect

from rdial.utils import AttrDict


class AttrDictTest(TestCase):
    def setUp(self):
        self.ad = AttrDict(carrots=3, snacks=0)

    def test_base(self):
        expect(self.ad).isinstance(dict)

        expect(self.ad['carrots']) == 3
        expect(self.ad['snacks']) == 0

        expect(sorted(self.ad.keys())) == ['carrots', 'snacks']

    def test___contains__(self):
        expect(self.ad).contains('carrots')
        expect(self.ad).does_not_contain('prizes')

    def test___getattr__(self):
        expect(self.ad.carrots) == 3
        expect(self.ad.snacks) == 0

    def test___setattr__(self):
        self.ad.carrots, self.ad.snacks = 0, 3
        expect(self.ad.carrots) == 0
        expect(self.ad.snacks) == 3

    def test___delattr__(self):
        expect(self.ad).contains('carrots')
        del self.ad['carrots']
        expect(self.ad).does_not_contain('carrots')



#
# coding=utf-8
"""test_datetime - Test ISO-8601 handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import datetime

from expecter import expect
from nose2.tools import params

from rdial.utils import (format_datetime, parse_datetime, utc)


@params(
    ('2011-05-04T08:00:00Z', datetime(2011, 5, 4, 8, 0, tzinfo=utc)),
    ('2011-05-04T09:15:00Z', datetime(2011, 5, 4, 9, 15, tzinfo=utc)),
    ('2011-05-04T10:15:00+0100', datetime(2011, 5, 4, 9, 15, tzinfo=utc)),
)
def test_parse_datetime(string, expected):
    expect(parse_datetime(string)) == expected


def test_parse_datetime_via_date_command(string, expected):
    now = datetime.datetime.now() - datetime.timedelta(minutes=40)
    expect(parse_datetime(now.isoformat())) == now


@params(
    (datetime(2011, 5, 4, 8, 0, tzinfo=utc), '2011-05-04T08:00:00Z'),
    (datetime(2011, 5, 4, 9, 15, tzinfo=utc), '2011-05-04T09:15:00Z'),
)
def test_format_datetime(dt, expected):
    expect(format_datetime(dt)) == expected



#
# coding=utf-8
"""test_delta - Test ISO-8601 delta handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import timedelta

from expecter import expect
from nose2.tools import params

from rdial.utils import (format_delta, parse_delta)


@params(
    ('PT04H30M21S', timedelta(hours=4, minutes=30, seconds=21)),
    ('PT00H12M01S', timedelta(minutes=12, seconds=1)),
)
def test_parse_duration(string, expected):
    expect(parse_delta(string)) == expected


@params(
    (timedelta(hours=4, minutes=30, seconds=21), 'PT04H30M21S'),
    (timedelta(minutes=12, seconds=1), 'PT12M01S'),
)
def test_format_duration(delta, expected):
    expect(format_delta(delta)) == expected


def test_parse_null_duration():
    expect(parse_delta('')) == timedelta()


def test_format_zero_duration():
    expect(format_delta(timedelta())) == ''


@params(
    ('PT04H', timedelta(hours=4)),
    ('PT04H30M', timedelta(hours=4, minutes=30)),
    ('PT30M', timedelta(minutes=30)),
    ('PT04H21S', timedelta(hours=4, seconds=21)),
    ('PT4H', timedelta(hours=4)),
)
def test_parse_partially_defined_durations(string, expected):
    expect(parse_delta(string)) == expected


@params(
    ('P3DT04H', timedelta(days=3, hours=4)),
    ('P3D', timedelta(days=3)),
)
def test_parse_durations_with_days(string, expected):
    expect(parse_delta(string)) == expected


@params(
    (timedelta(days=3, hours=4), 'P3DT04H'),
    (timedelta(days=3), 'P3D'),
    (timedelta(days=2, hours=22), 'P2DT22H'),
)
def test_format_durations_with_days(delta, expected):
    expect(format_delta(delta)) == expected


@params(
    (timedelta(hours=4), 'PT04H'),
    (timedelta(hours=4, minutes=30), 'PT04H30M'),
    (timedelta(minutes=30), 'PT30M'),
    (timedelta(hours=4, seconds=21), 'PT04H21S'),
)
def test_format_partially_defined_durations(delta, expected):
    expect(format_delta(delta)) == expected



#
# coding=utf-8
"""test_event_filter - Test event filter handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from expecter import expect
from nose2.tools import params

from rdial.events import Events


def test_fetch_events_for_task():
    events = Events.read('tests/data/test', write_cache=False)
    expect(len(events.for_task(task='task2'))) == 1


@params(
    ({'year': 2011, }, 2),
    ({'year': 2011, 'month': 1}, 1),
    ({'year': 2011, 'month': 3, 'day': 1}, 1),
    ({'year': 2011, 'month': 3, 'day': 31}, 0),
)
def test_fetch_events_for_date(date, expected):
    events = Events.read('tests/data/date_filtering', write_cache=False)
    expect(len(events.for_date(**date))) == expected


def test_fetch_events_for_week():
    events = Events.read('tests/data/date_filtering', write_cache=False)
    expect(len(events.for_week(year=2011, week=9))) == 1



#
# coding=utf-8
"""test_cmdline - Test command line functionality"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from click import BadParameter
from expecter import expect
from nose2.tools import params

from rdial.cmdline import (StartTimeParamType, TaskNameParamType)


@params(
    ('valid_name', True),
    ('also-valid-name', True),
    ('.invalid_name', BadParameter),
    ('valid.name', True),
    ('invalid/name', BadParameter),
)
def test_task_name_validity(string, expected):
    p = TaskNameParamType()
    if expected is True:
        p.convert(string, None, None) == string
    else:
        with expect.raises(expected):
            p.convert(string, None, None)


@params(
    ('yesterday', True),
    ('', True),
    ('2011-05-04T09:15:00Z', True),
    ('AB1 time', BadParameter),
)
def test_start_time_validity(string, expected):
    p = StartTimeParamType()
    if expected is True:
        p.convert(string, None, None) == string
    else:
        with expect.raises(expected):
            p.convert(string, None, None)



#
# coding=utf-8
"""utils - Utility functions for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# pylint: disable-msg=C0121

import datetime
import functools
import os
import re
import subprocess

import ciso8601
import click
import configobj

from . import compat


# Set up informational message functions
def _colourise(text, colour):
    """Colour text, if possible.

    :param str text: Text to colourise
    :param str colour: Colour to display text in
    :rtype: :obj:`str`
    :return str: Bright colourised text, if possible
    """
    return click.termui.secho(text, fg=colour, bold=True)


def success(text):
    """Pretty print a success message.

    :param str text: Text to format
    :rtype: :obj:`str`
    :return: Bright green text, if possible
    """
    return _colourise(text, 'green')


def fail(text):
    """Pretty print a failure message.

    :param str text: Text to format
    :rtype: :obj:`str`
    :return: Bright red text, if possible
    """
    return _colourise(text, 'red')


def warn(text):
    """Pretty print a warning message.

    :param str text: Text to format
    :rtype: ``str``
    :return: Bright yellow text, if possible
    """
    return _colourise(text, 'yellow')


class RdialError(ValueError):

    """Generic exception for rdial."""

    if not compat.PY2:
        @property
        def message(self):
            """Compatibility hack for Python 3."""
            return self.args[0]


class AttrDict(dict):

    """Dictionary with attribute access.

    .. seealso:: :obj:`dict`
    """

    def __contains__(self, key):
        """Check for item membership

        :param object key: Key to test for
        :rtype: :obj:`bool`
        """
        return hasattr(self, key) or super(AttrDict, self).__contains__(key)

    def __getattr__(self, key):
        """Support item access via dot notation

        :param object key: Key to fetch
        """
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        """Support item assignment via dot notation

        :param object key: Key to set value for
        :param object value: Value to set key to
        """
        try:
            self[key] = value
        except:
            raise AttributeError(key)

    def __delattr__(self, key):
        """Support item deletion via dot notation

        :param object key: Key to delete
        """
        try:
            del self[key]
        except KeyError:
            raise AttributeError(key)


class UTC(datetime.tzinfo):

    """UTC timezone object."""

    def __repr__(self):
        return 'UTC()'

    # pylint: disable-msg=W0613
    def utcoffset(self, datetime_):
        return datetime.timedelta(0)

    def dst(self, datetime_):
        return datetime.timedelta(0)

    def tzname(self, datetime_):
        return 'UTC'
    # pylint: enable-msg=W0613

utc = UTC()


def parse_delta(string):
    """Parse ISO-8601 duration string.

    :param str string: Duration string to parse
    :rtype: :obj:`datetime.timedelta`
    :return: Parsed delta object
    """
    if not string:
        return datetime.timedelta(0)
    match = re.match("""
        P
        ((?P<days>\d+)D)?
        T?
        ((?P<hours>\d{1,2})H)?
        ((?P<minutes>\d{1,2})M)?
        ((?P<seconds>\d{1,2})?(\.(?P<microseconds>\d+)S)?)
    """, string, re.VERBOSE)
    match_dict = dict((k, int(v) if v else 0)
                      for k, v in match.groupdict().items())
    return datetime.timedelta(**match_dict)  # pylint: disable-msg=W0142


def format_delta(timedelta_):
    """Format ISO-8601 duration string.

    :param datetime.timedelta timedelta_: Duration to process
    :rtype: :obj:`str`
    :return: ISO-8601 representation of duration
    """
    if timedelta_ == datetime.timedelta(0):
        return ''
    days = '%dD' % timedelta_.days if timedelta_.days else ''
    hours, minutes = divmod(timedelta_.seconds, 3600)
    minutes, seconds = divmod(minutes, 60)
    hours = '%02dH' % hours if hours else ''
    minutes = '%02dM' % minutes if minutes else ''
    seconds = '%02dS' % seconds if seconds else ''
    return 'P%s%s%s%s%s' % (days, 'T' if hours or minutes or seconds else '',
                            hours, minutes, seconds)


def parse_datetime(string):
    """Parse datetime string.

    :param str string: Datetime string to parse
    :rtype: :obj:`datetime.datetime`
    :return: Parsed datetime object
    """
    if not string:
        datetime_ = utcnow()
    else:
        datetime_ = ciso8601.parse_datetime(string)
        if not datetime_:
            try:
                output = check_output(['date', '--utc', '--iso-8601=seconds',
                                       '-d', string])
                datetime_ = ciso8601.parse_datetime(output.strip())
            except subprocess.CalledProcessError:
                pass
        if not datetime_:
            raise ValueError('Unable to parse timestamp %r' % string)
    return datetime_


def format_datetime(datetime_):
    """Format ISO-8601 datetime string.

    :param datetime.datetime datetime_: Datetime to process
    :rtype: str
    :return: ISO-8601 compatible string
    """
    # Can't call isoformat method as it uses the +00:00 form
    return datetime_.strftime('%Y-%m-%dT%H:%M:%SZ')


def iso_week_to_date(year, week):
    """Generate date range for a given ISO-8601 week.

    ISO-8601 defines a week as Monday to Sunday, with the first week of a year
    being the first week containing a Thursday.

    :param int year: Year to process
    :param int week: Week number to process
    :rtype: :obj:`tuple` of :obj:`datetime.date`
    :return: Date range objects for given week
    """
    bound = datetime.date(year, 1, 4)
    iso_start = bound - datetime.timedelta(days=bound.isocalendar()[2] - 1)
    start = iso_start + datetime.timedelta(weeks=week - 1)
    end = start + datetime.timedelta(weeks=1)
    return start, end


def utcnow():
    """Wrapper for producing timezone aware current timestamp.

    :rtype: obj:`datetime.datetime`
    :return: Current date and time, in UTC

    """
    return datetime.datetime.utcnow().replace(tzinfo=utc)


def check_output(args, **kwargs):
    """Simple check_output implementation for Python 2.6 compatibility.

    ..note:: This hides stderr, unlike the normal check_output function.

    :param list args: Command and arguments to call
    :rtype: ``str``
    :return: Command output
    :raise subprocess.CalledProcessError: If command execution fails
    """
    try:
        output = subprocess.check_output(args, stderr=subprocess.PIPE,
                                         **kwargs)
    except AttributeError:
        process = subprocess.Popen(args, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            raise subprocess.CalledProcessError(retcode, args[0])
    if not compat.PY2:
        output = output.decode()
    return output


def read_config(user_config=None):
    """Read configuration data.

    :type str user_config: User defined config file
    :rtype: configobj.ConfigObj
    :return: Parsed configuration data
    """
    configs = [os.path.dirname(__file__) + '/config', ]
    for s in os.getenv('XDG_CONFIG_DIRS', '/etc/xdg').split(':'):
        p = s + '/rdial/config'
        if os.path.isfile(p):
            configs.append(p)
    configs.append(xdg_config_location() + '/config')
    configs.append(os.path.abspath('.rdialrc'))
    if user_config:
        configs.append(user_config)
    # Prime config with dynamic key
    lines = ['xdg_data_location = %r' % xdg_data_location(), ]
    for file in configs:
        if os.path.isfile(file):
            lines.extend(compat.open(file, encoding='utf-8').readlines())
    return configobj.ConfigObj(lines)


def write_current(f):
    """Decorator to write ``current`` file on function exit.

    :seealso: :doc:`/taskbars`

    :rtype: :obj:`function`
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        globs = args[0]
        f(*args, **kwargs)
        open('%s/.current' % globs['directory'], 'w').write(kwargs['task'])
    return wrapper


def remove_current(f):
    """Decorator to remove ``current`` file on function exit.

    :seealso: :doc:`/taskbars`

    :rtype: :obj:`function`
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        globs = args[0]
        f(*args, **kwargs)
        if os.path.isfile('%s/.current' % globs['directory']):
            os.unlink('%s/.current' % globs['directory'])
    return wrapper


def newer(file, reference):
    """Check whether given file is newer than reference file.

    :param str file: File to check
    :param str reference: file to test against
    :rtype: :obj:`bool`
    :return: True if ``reference`` is newer than ``reference``
    """
    return os.stat(file).st_mtime > os.stat(reference).st_mtime


def xdg_cache_location():
    """Return a cache location honouring $XDG_CACHE_HOME.

    :rtype: :obj:`str`
    """
    user_dir = os.getenv('XDG_CACHE_HOME',
                         os.path.join(os.getenv('HOME', '/'), '.cache'))
    return os.path.join(user_dir, 'rdial')


def xdg_config_location():
    """Return a config location honouring $XDG_CONFIG_HOME.

    :rtype: :obj:`str`
    """
    user_dir = os.getenv('XDG_CONFIG_HOME',
                         os.path.join(os.getenv('HOME', '/'), '.config'))
    return os.path.join(user_dir, 'rdial')


def xdg_data_location():
    """Return a data location honouring $XDG_DATA_HOME.

    :rtype: :obj:`str`
    """
    user_dir = os.getenv('XDG_DATA_HOME', os.path.join(os.getenv('HOME', '/'),
                         '.local/share'))
    return os.path.join(user_dir, 'rdial')



#
# coding=utf-8
"""test_event - Test database handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from datetime import (datetime, timedelta)
from filecmp import dircmp
from os.path import abspath

from click.testing import CliRunner
from expecter import expect
from nose2.tools import params

from rdial.events import (Event, Events)
from rdial.utils import (parse_datetime, parse_delta, utc)


@params(
    ('test', None, None, None),
    ('test', '2013-02-26T19:45:14Z', None, None),
    ('test', datetime(2013, 2, 26, 19, 45, 14, tzinfo=utc), None,
     None),
    ('test', '2013-02-26T19:45:14Z', 'PT8M19.770869S', None),
    ('test', '2013-02-26T19:45:14Z', timedelta(minutes=8, seconds=19.770869),
     None),
    ('test', '2013-02-26T19:45:14Z', 'PT8M19.770869S', 'stopped'),
)
def test_event_creation(task, start, delta, message):
    e = Event(task, start, delta, message)
    expect(e.task) == task
    if isinstance(start, datetime):
        expect(e.start) == start
    elif start:
        expect(e.start) == parse_datetime(start)
    else:
        # Special case to ignore comparison against utcnow()
        pass
    if isinstance(delta, timedelta):
        expect(e.delta) == delta
    else:
        expect(e.delta) == parse_delta(delta)
    expect(e.message) == message


@params(
    ('test', 3),
    ('date_filtering', 3),
    ('test_not_running', 3),
)
def test_read_datebase(database, events):
    evs = Events.read('tests/data/' + database, write_cache=False)
    expect(len(evs)) == events


@params(
    (0, 'task', datetime(2011, 5, 4, 8, tzinfo=utc),
     timedelta(hours=1)),
    (1, 'task2', datetime(2011, 5, 4, 9, 15, tzinfo=utc),
     timedelta(minutes=15)),
    (2, 'task', datetime(2011, 5, 4, 9, 30, tzinfo=utc), timedelta()),
)
def test_check_events(n, task, start, delta):
    # FIXME: Clean-ish way to perform check, with the caveat that it parses the
    # database on each entry.  Need a better solution.
    events = Events.read('tests/data/test', write_cache=False)
    expect(events[n].task) == task
    expect(events[n].start) == start
    expect(events[n].delta) == delta


def test_write_database():
    runner = CliRunner()
    in_dir = abspath('tests/data/test')
    events = Events.read(in_dir, write_cache=False)
    events._dirty = events.tasks()
    with runner.isolated_filesystem() as tempdir:
        events.write(tempdir)
        comp = dircmp(in_dir, tempdir)
        expect(comp.diff_files) == []
        expect(comp.left_only) == []
        expect(comp.right_only) == []
        expect(comp.funny_files) == []


def test_store_messages_with_events():
    events = Events.read('tests/data/test', write_cache=False)
    expect(events.last().message) == 'finished'


def test_non_existing_database():
    expect(Events()) == Events.read('I_NEVER_EXIST', write_cache=False)



#
# coding=utf-8
"""cmdline - Command line functionality for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# pylint: disable-msg=C0121

import datetime
import operator
import os
import shlex
import subprocess

import click
import tabulate

from .events import (Events, TaskNotRunningError, TaskRunningError)
from .i18n import (_, N_)
from . import _version
from . import utils


class TaskNameParamType(click.ParamType):
    """Task name parameter handler."""
    name = 'taskname'

    def convert(self, value, param, ctx):
        """Check given task name is valid.

        :param str value: Value given to flag
        :param click.Argument param: Parameter being processed
        :param click.Context ctx: Current command context
        :rtype: :obj:`str`
        :return: Valid task name
        """
        if value.startswith('.') or '/' in value or '\000' in value:
            self.fail(_('%r is not a valid task name') % value)
        return value


class StartTimeParamType(click.ParamType):
    """Start time parameter handler."""
    name = 'time'

    def convert(self, value, param, ctx):
        """Check given start time is valid.

        :param str value: Value given to flag
        :param click.Argument param: Parameter being processed
        :param click.Context ctx: Current command context
        :rtype: :obj:`datetime.datetime`
        :return: Valid start time
        """
        try:
            value = utils.parse_datetime(value)
        except ValueError:
            self.fail(_('%r is not a valid ISO-8601 time string') % value)
        return value


def task_from_dir(ctx, param, value):
    """Override task name default using name of current directory.

    :param click.Context ctx: Current command context
    :param click.Argument param: Parameter being processed
    :param bool value: True if flag given
    """
    if value:
        param = [p for p in ctx.command.params if p.name == 'task'][0]
        param.default = os.path.basename(os.path.abspath(os.curdir))


def get_stop_message(current, edit=False):
    """Interactively fetch stop message.

    :param events.Event current: Current task
    :param bool edit: Whether to edit existing message
    :rtype: :obj:`str`
    :return: Message to use
    """
    marker = _('# Text below here ignored\n')
    task_message = _("# Task `%s' started %s") % (current.task,
                                                  current.start.humanize())
    template = "%s\n%s%s" % (current.message, marker, task_message)
    message = click.edit(template, require_save=not edit)
    if message is None:
        message = ''
    else:
        message = message.split(marker, 1)[0].rstrip('\n')
    return message


@click.group(help=_('Simple time tracking for simple people.'),
             epilog=_('Please report bugs to '
                      'https://github.com/JNRowe/rdial/issues'))
@click.version_option(_version.dotted)
@click.option('-d', '--directory', envvar='RDIAL_DIRECTORY', metavar='DIR',
              help=_('Directory to read/write to.'))
@click.option('--backup/--no-backup', envvar='RDIAL_BACKUP',
              help=_('Do not write data file backups.'))
@click.option('--cache/--no-cache', envvar='RDIAL_CACHE',
              help=_('Do not write cache files.'))
@click.option('--config', envvar='RDIAL_CONFIG', type=click.File(),
              help=_('File to read configuration data from.'))
@click.option('-i', '--interactive/--no-interactive',
              envvar='RDIAL_INTERACTIVE',
              help=_('Support interactive message editing.'))
@click.pass_context
def cli(ctx, directory, backup, cache, config, interactive):
    """Main command entry point.

    :param click.Context ctx: Current command context
    :param str directory: Location to store event data
    :param bool backup: Whether to create backup files
    :param bool cache: Whether to create cache files
    :param str config: Location of config file
    :param bool interactive: Whether to support interactive message editing
    """
    cfg = utils.read_config(config)
    base = cfg['rdial']

    if 'color' in base:
        base['colour'] = base['color']
    if not base.as_bool('colour') or os.getenv('NO_COLOUR') \
            or os.getenv('NO_COLOR'):
        utils._colourise = lambda s, colour: s

    ctx.default_map = {}
    for name in ctx.command.commands:
        if name in cfg.sections:
            d = {}
            for k in cfg[name]:
                try:
                    d[k] = cfg[name].as_bool(k)
                except ValueError:
                    d[k] = cfg[name][k]
            ctx.default_map[name] = d

    ctx.obj = utils.AttrDict(
        backup=backup or base.as_bool('backup'),
        cache=cache or base.as_bool('cache'),
        config=cfg,
        directory=directory or base['directory'],
        interactive=interactive or base.as_bool('interactive'),
    )


def filter_events(globs, task=None, duration=None):
    """Filter events for report processing.

    :param dict globs: Global options object
    :param str task: Task name to filter on
    :param str duration: Time window to filter on
    :rtype: :obj:`rdial.events.Events`
    :return: Events matching specified criteria
    """
    events = Events.read(globs.directory, write_cache=globs.cache)
    if task:
        events = events.for_task(task)
    if not duration == 'all':
        if duration == 'week':
            today = datetime.date.today()
            events = events.for_week(*today.isocalendar()[:2])
        else:
            year, month, day = datetime.date.today().timetuple()[:3]
            if duration == 'month':
                day = None
            elif duration == 'year':
                month = None
                day = None
            events = events.for_date(year, month, day)
    return events


@cli.command(help=_('Check storage consistency.'))
@click.pass_obj
@click.pass_context
def fsck(ctx, globs):
    """Check storage consistency.

    :param click.Context ctx: Current command context
    :param dict globs: Global options object
    """
    warnings = 0
    with Events.context(globs.directory, globs.backup, globs.cache) as events:
        last = events[0]
        for event in events[1:]:
            if not last.start + last.delta <= event.start:
                warnings += 1
                utils.fail(_('Overlap:'))
                utils.warn('  %r' % last)
                click.echo('  %r' % event)
            last = event
    if warnings:
        ctx.exit(warnings)


@cli.command(help=_('Start task.'))
@click.option('-x', '--from-dir', is_flag=True, expose_value=False,
              is_eager=True, callback=task_from_dir,
              help=_('Use directory name as task name.'))
@click.argument('task', default='default', envvar='RDIAL_TASK',
                required=False, type=TaskNameParamType())
@click.option('-n', '--new', is_flag=True, help=_('Start a new task.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@click.pass_obj
@utils.write_current
def start(globs, task, new, time):
    """Start task.

    :param dict globs: Global options object
    :param str task: Task name to operate on
    :param bool new: Create a new task
    :param datetime.datetime time: Task start time
    """
    with Events.context(globs.directory, globs.backup, globs.cache) as events:
        events.start(task, new, time)


@cli.command(help=_('Stop task.'))
@click.option('-m', '--message', help=_('Closing message.'))
@click.option('-F', '--file', type=click.File(),
              help=_('Read closing message from file.'))
@click.option('--amend', is_flag=True, help=_('Amend previous stop entry.'))
@click.pass_obj
@utils.remove_current
def stop(globs, message, file, amend):
    """Stop task.

    :param dict globs: Global options object
    :param str message: Message to assign to event
    :param str file: Filename to read message from
    :param bool amend: Amend a previously stopped event
    """
    if file:
        message = file.read()
    with Events.context(globs.directory, globs.backup, globs.cache) as events:
        last = events.last()
        running = last.running()
        if amend and running:
            raise TaskRunningError(_("Can't amend running task %s!")
                                   % last.task)
        elif not amend and not running:
            raise TaskNotRunningError(_('No task running!'))
        if amend and not message:
            message = last.message
        if globs.interactive and not message:
            get_stop_message(last, edit=amend)
        events.stop(message, force=amend)
    event = events.last()
    click.echo(_('Task %s running for %s') % (event.task,
                                              str(event.delta).split('.')[0]))


@cli.command(help=_('Switch to another task.'))
@click.option('-x', '--from-dir', is_flag=True, expose_value=False,
              is_eager=True, callback=task_from_dir,
              help=_('Use directory name as task name.'))
@click.argument('task', default='default', envvar='RDIAL_TASK',
                required=False, type=TaskNameParamType())
@click.option('-n', '--new', is_flag=True, help=_('Start a new task.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@click.option('-m', '--message',
              help=_('Closing message for current task.'))
@click.option('-F', '--file', type=click.File(),
              help=_('Read closing message for current task from file.'))
@click.pass_obj
@utils.write_current
def switch(globs, task, new, time, message, file):
    """Complete last task and start new one.

    :param dict globs: Global options object
    :param str task: Task name to operate on
    :param bool new: Create a new task
    :param datetime.datetime time: Task start time
    :param str message: Message to assign to event
    :param str file: Filename to read message from
    """
    if file:
        message = file.read()
    with Events.context(globs.directory, globs.backup, globs.cache) as events:
        event = events.last()
        if time and time < event.start:
            raise TaskNotRunningError(_("Can't specify a start time before "
                                        "current task started!"))
        if not event.running():
            raise TaskNotRunningError(_('No task running!'))
        if new or task in events.tasks():
            if globs.interactive and not message:
                get_stop_message(event)
            # This is dirty, but we kick on to Events.start() to save
            # duplication of error handling for task names
            events.stop(message)
        events.last().delta = time - event.start
        events.start(task, new, time)
    click.echo(_('Task %s running for %s') % (event.task,
                                              str(event.delta).split('.')[0]))


@cli.command(help=_('Run command with timer.'))
@click.option('-x', '--from-dir', is_flag=True, expose_value=False,
              is_eager=True, callback=task_from_dir,
              help=_('Use directory name as task name.'))
@click.argument('task', default='default', envvar='RDIAL_TASK',
                required=False, type=TaskNameParamType())
@click.option('-n', '--new', is_flag=True, help=_('Start a new task.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@click.option('-m', '--message', help=_('Closing message.'))
@click.option('-F', '--file', type=click.File(),
              help=_('Read closing message from file.'))
@click.option('-c', '--command', help=_('Command to run.'))
@click.pass_obj
def run(globs, task, new, time, message, file, command):
    """Run timed command.

    :param dict globs: Global options object
    :param str task: Task name to operate on
    :param bool new: Create a new task
    :param datetime.datetime time: Task start time
    :param str message: Message to assign to event
    :param str file: Filename to read message from
    :param str command: Command to run
    """
    with Events.context(globs.directory, globs.backup, globs.cache) as events:
        if events.running():
            raise TaskRunningError(_('Task %s is already started!'
                                     % events.last().task))

        try:
            p = subprocess.Popen(command, shell=True)
        except OSError as e:
            raise utils.RdialError(e.strerror)

        events.start(task, new, time)
        open('%s/.current' % globs.directory, 'w').write(task)

        p.wait()

        if file:
            message = file.read()
        if globs.interactive and not message:
            get_stop_message(events.running())
        events.stop(message)
    event = events.last()
    click.echo(_('Task %s running for %s') % (event.task,
                                              str(event.delta).split('.')[0]))
    os.unlink('%s/.current' % globs.directory)
    if p.returncode != 0:
        raise OSError(p.returncode, _('Command failed'))


@cli.command(help=_('Run predefined command with timer.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@click.option('-m', '--message', help=_('Closing message.'))
@click.option('-F', '--file', type=click.File(),
              help=_('Read closing message from file.'))
@click.argument('wrapper', default='default')
@click.pass_obj
@click.pass_context
def wrapper(ctx, globs, time, message, file, wrapper):
    """Run predefined timed command.

    :param click.Context ctx: Click context object
    :param dict globs: Global options object
    :param datetime.datetime time: Task start time
    :param str message: Message to assign to event
    :param str file: Filename to read message from
    :param str wrapper: Run wrapper to execute
    """
    if not 'run wrappers' in globs.config:
        raise ValueError(_('No %r section in config') % 'run wrappers')
    try:
        command = globs.config['run wrappers'][wrapper]
    except KeyError:
        raise ValueError(_('No such wrapper %r') % wrapper)
    parser = ctx.parent.command.commands['run'].make_parser(ctx)
    args = {'time': time, 'message': message, 'file': file, 'new': False}
    args.update(parser.parse_args(shlex.split(command))[0])
    ctx.invoke(run, **args)


@cli.command(help=_('Report time tracking data.'))
@click.option('-x', '--from-dir', is_flag=True, expose_value=False,
              is_eager=True, callback=task_from_dir,
              help=_('Use directory name as task name.'))
@click.argument('task', default='default', envvar='RDIAL_TASK',
                required=False, type=TaskNameParamType())
@click.option('--human', is_flag=True,
              help=_('Produce human-readable output.'))
@click.option('-d', '--duration', default='all',
              type=click.Choice(['day', 'week', 'month', 'year', 'all']),
              help=_('Filter events for specified time period.'))
@click.option('-s', '--sort', default='task', envvar='RDIAL_SORT',
              type=click.Choice(['task', 'time']), help=_('Field to sort by.'))
@click.option('-r', '--reverse/--no-reverse', default=False,
              envvar='RDIAL_REVERSE', help=_('Reverse sort order.'))
@click.option('--style', default='simple', envvar='RDIAL_TABLE_STYLE',
              type=click.Choice(tabulate._table_formats.keys()),
              help=_('Table output style.'))
@click.pass_obj
def report(globs, task, human, duration, sort, reverse, style):
    """Report time tracking data.

    :param dict globs: Global options object
    :param str task: Task name to operate on
    :param bool human: Display short overview of data
    :param str duration: Time window to filter on
    :param str sort: Key to sort events on
    :param bool reverse: Reverse sort order
    :param str style: Table formatting style
    """
    if task == 'default':
        # Lazy way to remove duplicate argument definitions
        task = None
    events = filter_events(globs, task, duration)
    if human:
        click.echo(N_('%d event in query', '%d events in query', len(events))
                   % len(events))
        click.echo(_('Duration of events %s') % events.sum())
        click.echo(_('First entry started at %s') % events[0].start)
        click.echo(_('Last entry started at %s') % events[-1].start)
        dates = set(e.start.date() for e in events)
        click.echo(_('Events exist on %d dates') % len(dates))
    else:
        data = sorted(([t, str(events.for_task(t).sum())]
                       for t in events.tasks()),
                      key=operator.itemgetter(['task', 'time'].index(sort)),
                      reverse=reverse)
        click.echo_via_pager(tabulate.tabulate(data, ['task', 'time'],
                                               tablefmt=style))
    if events.running():
        current = events.last()
        click.echo(_("Task `%s' started %s")
                   % (current.task, utils.format_datetime(current.start)))


@cli.command(help=_('Display running task, if any.'))
@click.pass_obj
def running(globs):
    """Display running task, if any.

    :param dict globs: Global options object
    """
    events = Events.read(globs.directory, write_cache=globs.cache)
    if events.running():
        current = events.last()
        click.echo(_("Task `%s' started %s")
                   % (current.task,
                      str(utils.utcnow() - current.start).split('.')[0]))
    else:
        utils.warn(_('No task is running!'))


@cli.command(help=_('Display last event, if any.'))
@click.pass_obj
def last(globs):
    """Display last event, if any.

    :param dict globs: Global options object
    """
    events = Events.read(globs.directory, write_cache=globs.cache)
    event = events.last()
    if not events.running():
        click.echo(_('Last task %s, ran for %s') % (event.task, event.delta))
        if event.message:
            click.echo(repr(event.message))
    else:
        utils.warn(_('Task %s is still running') % event.task)


@cli.command(help=_('Generate ledger compatible data file.'))
@click.option('-x', '--from-dir', is_flag=True, expose_value=False,
              is_eager=True, callback=task_from_dir,
              help=_('Use directory name as task name.'))
@click.argument('task', default='default', envvar='RDIAL_TASK',
                required=False, type=TaskNameParamType())
@click.option('-d', '--duration', default='all',
              type=click.Choice(['day', 'week', 'month', 'year', 'all']),
              help=_('Filter events for specified time period.'))
@click.option('-r', '--rate', envvar='RDIAL_RATE', type=click.FLOAT,
              help=_('Hourly rate for task output.'))
@click.pass_obj
def ledger(globs, task, duration, rate):
    """Generate ledger compatible data file.

    :param dict globs: Global options object
    :param str task: Task name to operate on
    :param str duration: Time window to filter on
    :param str rate: Rate to assign hours in report
    """
    if task == 'default':
        # Lazy way to remove duplicate argument definitions
        task = None
    events = filter_events(globs, task, duration)
    lines = []
    if events.running():
        lines.append(_(';; Running event not included in output!'))
    for event in events:
        if not event.delta:
            continue
        end = event.start + event.delta
        # Can't use timedelta.total_seconds() as it was only added in 2.7
        seconds = event.delta.days * 86400 + event.delta.seconds
        hours = seconds / 3600.0
        lines.append('%s-%s' % (event.start.strftime('%Y-%m-%d * %H:%M'),
                                end.strftime('%H:%M')))
        lines.append('    (task:%s)  %.2fh%s'
                     % (event.task, hours, ' @ %s' % rate if rate else ''))
    if events.running():
        lines.append(_(';; Running event not included in output!'))
    click.echo_via_pager('\n'.join(lines))


def main():
    """Command entry point to handle errors.

    :rtype: :obj:`int`
    :return: Final exit code
    """
    try:
        cli()
        return 0
    except (ValueError, utils.RdialError) as error:
        utils.fail(error.message)
        return 2
    except OSError as error:
        return error.errno



#
# coding=utf-8
"""test_fsck - Test database fsck support"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from click.testing import CliRunner
from expecter import expect

from rdial.cmdline import cli


def test_fsck_overlap():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory=tests/data/test_fsck',
                                 '--no-cache', 'fsck'])
    expect(result.exit_code) == 1
    expect(result.output).contains('Overlap')
    expect(result.output).contains("'2011-05-04T09:15:00Z', 'PT35M'")



#
# coding=utf-8
"""test_xdg_basedir - Test XDG base directory support"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from expecter import expect
from mock import patch

from rdial.utils import (xdg_config_location, xdg_data_location)


def test_config_no_args():
    with patch.dict('os.environ', {'XDG_CONFIG_HOME': '~/.xdg/config'}):
        expect(xdg_config_location()) == '~/.xdg/config/rdial'


def test_config_no_home():
    with patch.dict('os.environ', clear=True):
        expect(xdg_config_location()) == '/.config/rdial'


def test_data_no_args():
    with patch.dict('os.environ', {'XDG_DATA_HOME': '~/.xdg/local'}):
        expect(xdg_data_location()) == '~/.xdg/local/rdial'


def test_data_no_home():
    with patch.dict('os.environ', clear=True):
        expect(xdg_data_location()) == '/.local/share/rdial'
